import unittest
import numpy as np
import rose

def relative_difference(a, b):
    return np.abs((a-b)/b)


class TestMNPotential(unittest.TestCase):
    
    def test_mn_potential(self):
        self.assertTrue(
            relative_difference(
                rose.MN_Potential.v_r(1.0, np.array([1.0, 1.0])),
                0.854184893887586) < 1e-12,
            msg=f'V_MN = {rose.MN_Potential.v_r(1.0, np.array([1.0, 1.0]))}')


if __name__ == '__main__':
    unittest.main()


import numpy as np

HBARC = 197.326
DEFAULT_RHO_MESH = np.linspace(1e-6, 8*np.pi, 2000)
DEFAULT_ANGLE_MESH = np.linspace(0.01, np.pi, 10)
ALPHA = 0.0072973525693 # 1.0/137.056
MASS_PION = np.sqrt(1/2) # 2.0 per Thomspon&Nuñes below Eq. (4.3.10) 1/fm


'''Defines a class to package the spin-orbit term.
'''

from typing import Callable
import numpy as np

class SpinOrbitTerm:
    def __init__(self,
        spin_orbit_potential: Callable[[float, np.array, float], float], #V_{SO}(r, theta, l•s)
        l_dot_s: float # l•s
    ):
        r'''Spin-orbit interaction

        Parameters:
            spin_orbit_potential (Callable[[float, ndarray, float],float]):
                coordinate-space, spin-orbit potential
            l_dot_s (float): $2\ell\cdot s$ matrix elements, $+\ell$ or $-\ell-1$
        
        Attributes:
            l_dot_s (float): $2\ell\cdot s$ matrix elements, $+\ell$ or $-\ell-1$
            spin_orbit_potential: (Callable[[float, ndarray, float],float]):
                coordinate-space, spin-orbit potential

        '''
        self.l_dot_s = l_dot_s
        self.spin_orbit_potential = lambda r, alpha: spin_orbit_potential(r, alpha, self.l_dot_s)


'''The Koning-Delaroche potential is a common optical potential for nuclear
scattering. It is provided here in simplified form specifically to address this
need.

See the [Koning-Delaroche
paper](https://www.sciencedirect.com/science/article/pii/S0375947402013210) for
details. Equation references are with respect to (w.r.t.) this paper.
'''

import numpy as np

from .interaction_eim import InteractionEIM
from .energized_interaction_eim import EnergizedInteractionEIM
from .constants import DEFAULT_RHO_MESH, MASS_PION

MAX_ARG = np.log(1/1e-16)
NUM_PARAMS = 15

def Vv(E, v1, v2, v3, v4, Ef):
    r'''energy-dependent, volume-central strength - real term, Eq. (7)
    '''
    return v1 * (1 - v2*(E-Ef) + v3*(E-Ef)**2 - v4*(E-Ef)**3)


def Wv(E, w1, w2, Ef):
    '''energy-dependent, volume-central strength - imaginary term, Eq. (7)
    '''
    return w1 * (E-Ef)**2 / ((E-Ef)**2 + w2**2)


def Wd(E, d1, d2, d3, Ef):
    '''energy-dependent, surface-central strength - imaginary term (no real
        term), Eq. (7)
    '''
    return d1 * (E-Ef)**2 / ((E-Ef)**2 + d3**2) * np.exp(-d2*(E-Ef))


def Vso(E, vso1, vso2, E_f):
    '''energy-dependent, spin-orbit strength --- real term, Eq. (7)'''
    return vso1 * np.exp(-vso2*(E-E_f))


def Wso(E, wso1, wso2, E_f):
    '''energy-dependent, spin-orbit strength --- imaginary term, Eq. (7)'''
    return wso1 * (E - E_f)**2 / ((E - E_f)**2 + wso2**2)


def woods_saxon(r, R, a):
    '''Woods-Saxon potential'''
    return 1/(1 + np.exp((r-R)/a))


def woods_saxon_safe(r, R, a):
    '''Woods-Saxon potential

     * avoids `exp` overflows

     '''
    x = (r-R)/a
    if isinstance(x, float):
        return 1/(1 + np.exp(x)) if x < MAX_ARG else 0
    else:
        ii = np.where(x <= MAX_ARG)[0]
        jj = np.where(x > MAX_ARG)[0]
        return np.hstack([1/(1 + np.exp(x[ii])), np.zeros(jj.size)])
    

def woods_saxon_prime(r, R, a):
    '''derivative of the Woods-Saxon potential w.r.t. $r$
    '''
    return -1/a * np.exp((r-R)/a) / (1 + np.exp((r-R)/a))**2


def woods_saxon_prime_safe(r, R, a):
    '''derivative of the Woods-Saxon potential w.r.t. $r$

    * avoids `exp` overflows

    '''
    x = (r-R)/a
    if isinstance(x, float):
        return -1/a * np.exp(x)/(1 + np.exp(x))**2 if x < MAX_ARG else 0
    else:
        ii = np.where(x <= MAX_ARG)[0]
        jj = np.where(x > MAX_ARG)[0]
        return np.hstack([-1/a * np.exp(x[ii])/(1 + np.exp(x[ii]))**2, np.zeros(jj.size)])


def KD(r, E, v1, v2, v3, v4, w1, w2, d1, d2, d3, Ef, Rv, av, Rd, ad):
    '''Koning-Delaroche without the spin-orbit terms - Eq. (1)'''
    return -Vv(E, v1, v2, v3, v4, Ef) * woods_saxon(r, Rv, av) - \
           1j*Wv(E, w1, w2, Ef) * woods_saxon(r, Rv, av) - \
           1j * (-4*ad) * Wd(E, d1, d2, d3, Ef) * woods_saxon_prime(r, Rd, ad)


def decompose_alpha(alpha):
    r'''Splits the parameter-space vector into non-spin-orbit and spin-orbit
    parameters.

    Parameters:
        alpha (ndarray): interaction parameters
    
    Returns:
        parameters (tuple): 2-tuple of non-spin-orbit (`parameters[0]`) and
            spin-orbit parameters (`parameters[1]`)

    '''
    vv, rv, av, wv, rwv, awv, wd, rd, ad, vso, rso, aso, wso, rwso, awso = alpha
    return (vv, rv, av, wv, rwv, awv, wd, rd, ad), (vso, rso, aso, wso, rwso, awso)

def KD_simple(r, alpha):
    r'''simplified Koning-Delaroche without the spin-orbit terms

    Take Eq. (1) and remove the energy dependence of the coefficients.
    '''
    vv, rv, av, wv, rwv, awv, wd, rd, ad = decompose_alpha(alpha)[0]
    return -vv * woods_saxon_safe(r, rv, av) - \
        1j*wv * woods_saxon_safe(r, rwv, awv) - \
        1j*(-4*ad)*wd * woods_saxon_prime_safe(r, rd, ad)


def KD_simple_so(r, alpha, lds):
    r'''simplified Koning-Delaroche spin-orbit terms

    Take Eq. (1) and remove the energy dependence of the coefficients.

    lds: l • s = 1/2 * (j(j+1) - l(l+1) - s(s+1))
    '''
    vso, rso, aso, wso, rwso, awso = decompose_alpha(alpha)[1]
    return lds * vso/MASS_PION**2/r * woods_saxon_prime_safe(r, rso, aso) + \
        1j*wso/MASS_PION**2/r*woods_saxon_prime_safe(r, rwso, awso)


class KoningDelaroche(InteractionEIM):
    r'''Koning-Delaroche potential (without energy-dependent strength
    coefficients) for arbitrary systems defined by `mu`, `energy`, `ell`, `Z_1`,
    and `Z_2`.
    '''
    def __init__(self,
        mu: float,
        ell: int,
        energy: float,
        training_info: np.array,
        n_basis: int = 8,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''Wraps the Koning-Delaroche potential into a `rose`-friendly class.
        Saves system-specific information.
        
        Parameters:
            mu (float): reduced mass of the 2-body system
            ell (int): angular momentum
            energy (float): center-of-mass, scattering energy
            training_info (ndarray): either (1) parameters bounds or (2) explicit training points

                If (1):
                This is a 2-column matrix. The first column are the lower
                bounds. The second are the upper bounds. Each row maps to a
                single parameter.

                If (2):
                This is an MxN matrix. N is the number of parameters. M is the
                number of samples.
            n_basis (int): number of basis states to use for EIM
            explicit_training (bool): True implies training_info case (2); False implies (1)
            n_train (int): how many training samples to use
            rho_mesh (ndarray):  $\rho$ (or $s$) grid values
            match_points (ndarray): $\rho$ points at which we demand the EIMed
                potential match the true potential

        Returns:
            instance (KoningDelaroche): instance of the class

        ''' 
        super().__init__(
            KD_simple_so, NUM_PARAMS, mu, ell, energy, training_info=training_info, Z_1=0, Z_2=0,
            is_complex=True, n_basis=n_basis,
            explicit_training=explicit_training, n_train=n_train,
            rho_mesh=rho_mesh, match_points=match_points
        )


class EnergizedKoningDelaroche(EnergizedInteractionEIM):
    def __init__(self,
        mu: float,
        ell: int,
        training_info: np.array,
        n_basis: int = 8,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''Wraps the Koning-Delaroche potential into a `rose`-friendly class.
        Saves system-specific information. Allows the user to emulate across
        energies.

        * **Does not (yet) support Coulomb.**
        
        Parameters:
            mu (float): reduced mass of the 2-body system
            ell (int): angular momentum
            training_info (ndarray): either (1) parameters bounds or (2) explicit training points

                If (1):
                This is a 2-column matrix. The first column are the lower
                bounds. The second are the upper bounds. Each row maps to a
                single parameter.

                If (2):
                This is an MxN matrix. N is the number of parameters. M is the
                number of samples.
            n_basis (int): number of basis states to use for EIM
            explicit_training (bool): True implies training_info case (2); False implies (1)
            n_train (int): how many training samples to use
            rho_mesh (ndarray):  $\rho$ (or $s$) grid values
            match_points (ndarray): $\rho$ points at which we demand the EIMed
                potential match the true potential
        
        Returns:
            instance (EnergizedKoningDelaroche): instance of the class
        ''' 
        super().__init__(
            KD_simple_so, NUM_PARAMS, mu, ell, training_info=training_info, Z_1=0, Z_2=0,
            is_complex=True, n_basis=n_basis,
            explicit_training=explicit_training, n_train=n_train,
            rho_mesh=rho_mesh, match_points=match_points
        )



import pickle
from typing import Callable

import numpy as np 
import numpy.typing as npt
from scipy.interpolate import interp1d
from mpmath import coulombf

from .constants import HBARC
from .schroedinger import SchroedingerEquation
from .energized_interaction_eim import EnergizedInteractionEIM

class Basis:
    '''Base class / template
    '''

    @classmethod
    def load(cls, filename):
        r'''Loads a previously saved Basis.
        
        '''
        with open(filename, 'rb') as f:
            basis = pickle.load(f)
        return basis


    def __init__(self,
        solver: SchroedingerEquation,
        theta_train: np.array,
        rho_mesh: np.array,
        n_basis: int,
        l: int
    ):
        r'''Builds a reduced basis.
        
        Parameters:
            solver (SchroedingerEquation): high-fidelity solver
            theta_train (ndarray): training space
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            l (int): orbital angular momentum
        
        Attributes:
            solver (SchroedingerEquation): high-fidelity solver
            theta_train (ndarray): training space
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            l (int): orbital angular momentum

        '''
        self.solver = solver
        self.theta_train = np.copy(theta_train)
        self.rho_mesh = np.copy(rho_mesh)
        self.n_basis = n_basis
        self.l = l
    

    def phi_hat(self, coefficients):
        r'''Emulated wave function.
        
        Every basis should know how to reconstruct hat{phi} from a set of
        coefficients. However, this is going to be different for each basis, so
        we will leave it up to the subclasses to implement this.

        Parameters:
            coefficients (ndarray): expansion coefficients
        
        Returns:
            phi_hat (ndarray): approximate wave function

        '''
        raise NotImplementedError
    

    def save(self, filename):
        '''Saves a basis to file.
        
        Parameters:
            filename (string): name of file

        '''
        with open(filename, 'wb') as f:
            pickle.dump(self, f)
    

class RelativeBasis(Basis):
    def __init__(self,
        solver: SchroedingerEquation,
        theta_train: np.array,
        rho_mesh: np.array,
        n_basis: int,
        l: int,
        use_svd: bool,
        phi_0_energy: float = None
    ):
        r'''Builds a "relative" reduced basis. This is the default choice.

        $$
        \phi_{\rm HF} \approx \hat{\phi} = \phi_0 + \sum_i c_i \tilde{\phi}_i~.
        $$
        
        Parameters:
            solver (SchroedingerEquation): high-fidelity solver
            theta_train (ndarray): training space
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            l (int): orbital angular momentum
            use_svd (bool): Use principal components for $\tilde{\phi}$?
            phi_0_energy (float): energy at which $\phi_0$ is calculated
        
        Attributes:
            solver (SchroedingerEquation): high-fidelity solver
            theta_train (ndarray): training space
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            l (int): orbital angular momentum
            phi_0 (ndarray): free solution (no interaction)
            all_vectors (ndarray): HF solutions minus the free solution (one for
                each column)
            pillars (ndarray): $\tilde{\phi}_i$
            singular_values (ndarray): singular values from SVD
            vectors (ndarray): copy of `pillars`

        '''

        super().__init__(solver, theta_train, rho_mesh, n_basis, l)

        if phi_0_energy:
            k = np.sqrt(2*self.solver.interaction.mu*phi_0_energy/HBARC)
            eta = self.solver.interaction.k_c / k
        else:
            if isinstance(self.solver.interaction, EnergizedInteractionEIM):
                # k_mean = np.sqrt(2*self.solver.interaction.mu*np.mean(theta_train[:, 0])/HBARC)
                # eta = self.solver.interaction.k_c / k_mean
                # Does not support Coulomb (yet).
                eta = 0.0
            else:
                # If the phi_0 energy is not specified, we're only going to work
                # with non-Coulombic systems (for now).
                eta = 0.0
        
        # Returns Bessel functions when eta = 0.
        self.phi_0 = np.array([coulombf(self.l, eta, rho) for rho in self.rho_mesh], dtype=np.float64)

        self.all_vectors = np.array([
            self.solver.phi(theta, self.rho_mesh, l) - self.phi_0 for theta in theta_train
        ]).T
        self.pillars = self.all_vectors.copy()

        if use_svd:
            U, S, _ = np.linalg.svd(self.all_vectors, full_matrices=False)
            self.singular_values = np.copy(S)
            self.pillars = np.copy(U)
        
        self.vectors = np.copy(self.pillars[:, :self.n_basis])
    

    def phi_hat(self, coefficients):
        r'''Emulated wave function.
        
        Parameters:
            coefficients (ndarray): expansion coefficients
        
        Returns:
            phi_hat (ndarray): approximate wave function

        '''
        return self.phi_0 + np.sum(coefficients * self.vectors, axis=1)


class CustomBasis(Basis):
    def __init__(self,
        solutions: np.array, # HF solutions, columns
        phi_0: np.array, # "offset", generates inhomogeneous term
        rho_mesh: np.array, # rho mesh; MUST BE EQUALLY SPACED POINTS!!!
        n_basis: int,
        ell: int, # angular momentum, l
        use_svd: bool
    ):
        r'''Builds a custom basis. Allows the user to supply their own.

        $$
        \phi_{\rm HF} \approx \hat{\phi} = \phi_0 + \sum_i c_i \tilde{\phi}_i~.
        $$
        
        Parameters:
            solutions (ndarray): HF solutions
            phi_0 (ndarray): free solution (no interaction)
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            ell (int): orbital angular momentum
            use_svd (bool): Use principal components for $\tilde{\phi}$?
        
        Attributes:
            solver (SchroedingerEquation): not specified or assumed at construction
            theta_train (ndarray): not specified or assumed at construction
            rho_mesh (ndarray): discrete $s=kr$ mesh points
            n_basis (int): number of states in the expansion
            ell (int): orbital angular momentum
            phi_0 (ndarray): free solution (no interaction)
            solutions (ndarray): HF solutions provided by the user
            pillars (ndarray): $\tilde{\phi}_i$
            singular_values (ndarray): singular values from SVD
            vectors (ndarray): copy of `pillars`
            phi_0_interp (interp1d): interpolating function for $\phi_0$
            vectors_interp (interp1d): interpolating functions for vectors (basis states)

        '''

        super().__init__(None, None, rho_mesh, n_basis, ell)

        self.solutions = solutions.copy()
        self.pillars = solutions.copy()
        self.rho_mesh = rho_mesh.copy()
        self.n_basis = n_basis
        # Energy and angular momentum are not actually used, but it may be
        # useful to store them here to help the user keep track of which
        # CustomBasis this is.
        self.ell = ell

        self.phi_0 = phi_0.copy()

        if use_svd:
            U, S, _ = np.linalg.svd(self.solutions - self.phi_0[:, np.newaxis], full_matrices=False)
            self.singular_values = S.copy()
            self.pillars = U.copy()
        else:
            self.singular_values = None
        
        self.vectors = self.pillars[:, :self.n_basis].copy()
        
        # interpolating functions
        # To extrapolate or not to extrapolate?
        self.phi_0_interp = interp1d(self.rho_mesh, self.phi_0, kind='cubic')
        self.vectors_interp = [interp1d(self.rho_mesh, row, kind='cubic') for row in self.vectors.T]
    

    def phi_hat(self, coefficients):
        r'''Emulated wave function.
        
        Parameters:
            coefficients (ndarray): expansion coefficients
        
        Returns:
            phi_hat (ndarray): approximate wave function

        '''
        return self.phi_0 + np.sum(coefficients * self.vectors, axis=1)


'''
Defines a benchmark class that computes:
    * phase shifts
    * partial wave amplitudes
    * scattering amplitude
    * wave functions: u(s)
    * mesh: s
'''
import numpy as np
import numpy.typing as npt
from scipy.special import eval_legendre

import rose


ANGLES = np.linspace(0, np.pi, 100)
COSTHETA = np.cos(ANGLES)

class BenchmarkData():
    '''
    Computes a set of quantities that we can compare to later.
    '''
    def __init__(self,
        schrodeq: rose.SchroedingerEquation,
        energy: float, # MeV, c.m.
        theta: npt.ArrayLike, # interaction parameters; not related to cos(theta)
        n: int = 2000, # number of point in r to calculate u(r)
        l_max: int = 3, # maximum angular momentum
        r_0: float = rose.schroedinger.DEFAULT_R_0, # fm, delta <- u(r_0)
        costheta: npt.ArrayLike = COSTHETA # angles at which P_l(cos(theta)) are calculated
    ):
        self.theta = np.copy(theta)

        k = np.sqrt(2*rose.MN_Potential.mu*energy/rose.constants.HBARC)
        r_mesh = np.linspace(rose.schroedinger.DEFAULT_R_MIN,
                             rose.schroedinger.DEFAULT_R_MAX, n)
        s_mesh = k * r_mesh
        s_endpts = np.array([s_mesh[0], s_mesh[-1]])
            
        self.l_max = l_max
        l_values = np.arange(self.l_max+1)
        self.costheta = np.copy(costheta)
        self.deltas_l = np.array(
            [schrodeq.delta(energy, theta, s_endpts, l, k*r_0) for l in l_values]
        )
        self.fs_l = np.array(
            [1/(k/np.tan(delta_l) - 1j*k) for delta_l in self.deltas_l]
        )
        Pl = np.array([eval_legendre(l, self.costheta) for l in l_values]).T
        self.scattering_amplitude = np.sum((2*l_values + 1.0) * self.fs_l * Pl, axis=1)

        sol = schrodeq.solve_se(energy, theta, s_endpts, 0)
        self.s = np.copy(s_mesh)
        phi = sol(self.s)[0, :]
        c = np.max(np.abs(phi)) # normalize to 1
        self.phi = phi/c


# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.9.5'  # type: str
__version_tuple__ = version_tuple = (0, 9, 5)  # type: tuple[int | str, ...]



import unittest
import numpy as np

import rose

# Values used in Pablo's notebook.
ENERGY = 50 # MeV
V0R = 200
V0S = -91.85
_, u_pg = np.loadtxt('u_mn_test.txt', unpack=True)
u_pg /= np.max(u_pg) # This is a very hacky way to handle normalization.

class TestSchrEq(unittest.TestCase):
    # def test_schrodinger_equation(self):
    #     se = roqse.SchroedingerEquation(roqse.MN_Potential)
    #     solution = se.solve_se(ENERGY, np.array([V0R, V0S]), r_min=0.01)
    #     s = solution[:, 0]
    #     u = solution[:, 1] / np.max(solution[:, 1])
    #     self.assertTrue(
    #         np.linalg.norm(u - u_pg) < 0.1
    #     )
    

    def test_energy(self):
        '''
        Tests the correct handling of energy-dependent interactions. Does not
        use emulation.
        Compares the HF solutions from two interactions -- one that is
        energy-dependent and one that is not.
        '''
        mu = 1.0
        energy = 50.0
        ell = 0
        z = 0
        rho = rose.constants.DEFAULT_RHO_MESH
        potential = lambda r, theta: theta[0] * rose.koning_delaroche.woods_saxon(r, theta[1], theta[2]) 

        interaction = rose.Interaction(potential, 3, mu, energy, z, z)
        se1 = rose.SchroedingerEquation(interaction)

        theta1 = np.array([-10.0, 3.0, 1.0])
        theta2 = np.hstack((energy, theta1))

        train = np.array([
            [49.0, 51.0],
            [-15.0, -5.0],
            [2.0, 4.0],
            [0.8, 1.2]
        ])
        energized_interaction = rose.EnergizedInteractionEIM(
            potential, 3, mu, ell, train, Z_1=z, Z_2=z, n_train=20
        )
        se2 = rose.SchroedingerEquation(energized_interaction)

        phi1 = se1.phi(theta1, rho, ell)
        phi2 = se2.phi(theta2, rho, ell)
        norm_diff = np.linalg.norm(phi1 - phi2)

        self.assertTrue(
            norm_diff < 1e-16,
            msg=f'norm(difference) = {norm_diff:.2e}'
        )
    

    def test_emulation(self):
        AMU = 931.5
        A = 40
        mu = A / (A + 1) * AMU
        energy = 13.659

        alphastar = np.array([51.9, 4.00134372,  0.75, 0.4, 4.5143365, 0.51, 8.3,
            4.5143365, 0.51, 6.2, 3.45415141, 0.75, 1.0, 3.45415141, 0.75])
        ntheta = alphastar.size

        train = np.load('train.npy')

        ell = 3
        interaction = rose.InteractionEIM(rose.koning_delaroche.KD_simple,
                                          ntheta, mu, energy, ell, train, is_complex=True,
                                          spin_orbit_term=rose.SpinOrbitTerm(rose.koning_delaroche.KD_simple_so, ell),
                                          n_basis=ntheta+16, explicit_training=True)

        y = interaction.tilde(interaction.s_mesh, alphastar)
        yp = interaction.tilde_emu(alphastar)

        norm_diff = np.linalg.norm(y-yp)
        self.assertTrue(
            norm_diff < 1e-4,
            msg=f'norm(difference) = {norm_diff:.2e}'
        )

if __name__ == '__main__':
    unittest.main()


'''
Useful utility functions that I don't want to clutter up other modules with.
'''
import numpy as np
import numpy.typing as npt
from scipy.sparse import diags, lil_matrix
from scipy.misc import derivative
from scipy.special import eval_legendre

def finite_difference_first_derivative(
    s_mesh: npt.ArrayLike,
    sparse: bool = False
):
    '''
    Computes a finite difference matrix that (when applied) represent the first
    derivative.
    '''
    dx = s_mesh[1] - s_mesh[0]
    assert np.all(np.abs(s_mesh[1:] - s_mesh[:-1] - dx) < 1e-14), '''
Spacing must be consistent throughout the entire mesh.
    '''
    n = s_mesh.size
    coefficients = np.array([1, -8, 8, -1]) / (12*dx)
    if sparse:
        D1 = lil_matrix(diags(coefficients, [-2, -1, 1, 2], shape=(n, n)))
    else:
        D1 = lil_matrix(diags(coefficients, [-2, -1, 1, 2], shape=(n, n))).toarray()

    # Use O(dx^2) forward difference approximations for the first 2 rows.
    D1[0, :5] = np.array([-3, 4, -1, 0, 0]) / (2*dx)
    D1[1, :5] = np.array([0, -3, 4, -1, 0]) / (2*dx)

    # Use O(dx^2) backward difference approximations for the last 2 rows.
    D1[-2, -5:] = np.array([0, 1, -4, 3, 0]) / (2*dx)
    D1[-1, -5:] = np.array([0, 0, 1, -4, 3]) / (2*dx)

    return D1



def finite_difference_second_derivative(
    s_mesh: npt.ArrayLike,
    sparse: bool = False
):
    '''
    Computes a finite difference matrix that represents the second derivative
    (w.r.t. s or rho) operator in coordinate space.
    '''
    dx = s_mesh[1] - s_mesh[0]
    assert np.all(np.abs(s_mesh[1:] - s_mesh[:-1] - dx) < 1e-14), '''
Spacing must be consistent throughout the entire mesh.
    '''
    n = s_mesh.size
    coefficients = np.array([-1, 16, -30, 16, -1]) / (12*dx**2)
    if sparse:
        D2 = lil_matrix(diags(coefficients, [-2, -1, 0, 1, 2], shape=(n, n)))
    else:
        D2 = diags(coefficients, [-2, -1, 0, 1, 2], shape=(n, n)).toarray()

    # Use O(dx^2) forward difference approximation for the first 2 rows.
    D2[0, :5] = np.array([2, -5, 4, -1, 0]) / dx**2
    D2[1, :5] = np.array([0, 2, -5, 4, -1]) / dx**2

    # Use O(dx^2) backward difference approximation for the last 2 rows.
    D2[-2, -5:] = np.array([-1, 4, -5, 2, 0]) / dx**2 
    D2[-1, -5:] = np.array([0, -1, 4, -5, 2]) / dx**2 

    return D2


def regular_inverse_r(r, r_c):
    if isinstance(r, float):
        return 1/(2*r_c) * (3 - (r/r_c)**2) if r < r_c else 1/r
    else:
        ii = np.where(r <= r_c)[0]
        jj = np.where(r > r_c)[0]
        return np.hstack([1/(2*r_c) * (3 - (r[ii]/r_c)**2), 1/r[jj]])


def regular_inverse_s(s, s_c):
    if isinstance(s, float) or isinstance(s, int):
        return 1/(2*s_c) * (3 - (s/s_c)**2) if s < s_c else 1/s
    else:
        ii = np.where(s <= s_c)[0]
        jj = np.where(s > s_c)[0]
        return np.hstack([1/(2*s_c) * (3 - (s[ii]/s_c)**2), 1/s[jj]])


def eval_assoc_legendre(n, x):
        if n == 0:
            return np.zeros(x.size)
        else:
            return -(1-x**2)**(1/2) * derivative(lambda z: eval_legendre(n, z), x, dx=1e-9)